# This Python file uses the following encoding: utf-8

from constants import *
from Media import *
from random import randint
import copy

# КЛАСС ИГРЫ
class Game:
    def __init__(self):   
        # размеры доски в клетках:
        self.cols = SIZE
        self.rows = SIZE
        # игровое поле:
        self.board = None
    
        # счёт игры:
        self.result = [0] * 3
    
        # число ходов:
        self.nMove = 0
        # игрок, делающий ход:
        self.player = PLAYER
        # победитель:
        self.winner = None
    
        # глубина поиска:
        self.depth = 2
      
        # состояние игры:
        self.flgGameOver = True

    
    # НОВАЯ ИГРА              
    def newGame(self):
        # готовимся к новой игре -->
        # создаём пустое поле:
        self.board = self.createArray(self.rows, self.cols)
        # обнуляем число ходов:
        self.nMove = 0    
        # кто делает первый ход:
        self.player = randint(PLAYER, COMPUTER)
        self.flgGameOver = False

     
    # СОЗДАЁМ ДВУМЕРНЫЙ СПИСОК
    def createArray(self, cols, rows):
        # игровое поле:
        return [[IEMPTY for row in range(rows)] for col in range(cols)]
        

    # РИСУЕМ ИГРОВОЕ ПОЛЕ
    def drawBoard(self):
        for row in range(self.rows):
            for col in range(self.cols):
                item = self.board[col][row]
                x = 103 + 100 * col
                y = 203 + 100 * row
                if item == IEMPTY: pass
                elif item == IPLAYER: image(Media.imgX, x, y)
                elif item == ICOMPUTER: image(Media.imgO, x, y)
                
            
    # ВОЗВРАЩАЕМ КООРДИНАТЫ ХОДА КОМПЬЮТЕРА
    def getMoveCoords(self):
        # получаем ход:
        comp_move = self.maximizePlay(self.board, self.depth)
        # координаты клетки:
        coords = { "col": comp_move[0], "row": comp_move[1] }
        #print('coords ' + str(coords["col"]) + ' ' + str(coords["row"]))
        #print('score = ' + str(comp_move[2]))
        return coords
    

    # ХОД КОМПЬЮТЕРА
    def maximizePlay(self, board, depth):
        # оцениваем позицию на доске:
        score = self.calcScore(board)
    
        # терминальная позиция:
        if (self.isTerminated(board, depth, score)):
            return [None, None, score]
    
        # максимальная оценка:
        # col, row, score
        max_score = [None, None, -WIN_SCORE + 1]
    
        # перебираем все клетки доски:
        for row in range(self.rows):
            for col in range(self.cols): 
                # не пустая клетка:
                if (board[col][row] != IEMPTY):
                    continue
    
                # создаём копию доски:
                new_board = self.boardCopy(board)
                # ставим на свободную клетку 
                # фишку Компьютера:
                new_board[col][row] = ICOMPUTER
    
                # ход Игрока:
                playerMove = self.minimizePlay(new_board, depth - 1)
    
                # запоминаем лучший ход:
                if (max_score[0] == None or playerMove[2] > max_score[2]):
                    max_score[0] = col
                    max_score[1] = row
                    max_score[2] = playerMove[2]

        #print('max_score = ' + str(max))
        return max_score
  

    # ХОД ИГРОКА
    def minimizePlay(self, board, depth):
        score = self.calcScore(board)

        if (self.isTerminated(board, depth, score)):
            return [None, None, score]

        # минимальная оценка:
        # col, row, score
        min_score = [None, None, WIN_SCORE - 1]

        for row in range(self.rows):
            for col in range(self.cols):
                # не пустая клетка:
                if (board[col][row] != IEMPTY):
                    continue
    
                new_board = self.boardCopy(board)
                new_board[col][row] = IPLAYER
    
                computerMove = self.maximizePlay(new_board, depth - 1)
    
                if (min_score[0] == None or computerMove[2] < min_score[2]):
                    min_score[0] = col
                    min_score[1] = row
                    min_score[2] = computerMove[2]
        return min_score
    
    
    def place(self, board, col, row, code):
        # Check if column valid
        # 1. not empty 2. not exceeding the board size
        if (board[col][row] == IEMPTY):
            board[col][row] = code
            return True
        else:
            return False
        
        
    # ВОЗВРАШАЕМ true,
    # ЕСЛИ ДОСТИГНУТВ ТЕРМИНАЛЬНАЯ
    # ПОЗИЦИЯ
    def isTerminated(self, board, depth, score):
        if (depth == 0 or \
            score == WIN_SCORE or \
            score == -WIN_SCORE or \
            self.isFull(board)):
                return True
        return False
    

    # ОЦЕНОЧНАЯ ФУНКЦИЯ
    def calcScore(self, board):
        # общая оценка:
        points = 0
    
        # считаем фишки в вертикальных рядах:
        score = 0
        for col in range(self.cols):
            score = self.scorePosition(board, col, 0, 0, 1)
            # выиграл Компьютер:
            if (score == WIN_SCORE):
                return WIN_SCORE
            # выиграл Игрок:
            if (score == -WIN_SCORE):
                return -WIN_SCORE;
            # промежуточная позиция -->
            # добавляем очки:
            points += score
    
        # считаем фишки в горизонтальных рядах:
        for row in range(self.rows):
            score = self.scorePosition(board, 0, row, 1, 0)
            if (score == WIN_SCORE):
                return WIN_SCORE
            if (score == -WIN_SCORE):
                return -WIN_SCORE
            points += score
    
        # считаем фишки в нисходящей диагонали:
        score = self.scorePosition(board, 0, 0, 1, 1)
        if (score == WIN_SCORE):
            return WIN_SCORE
        if (score == -WIN_SCORE):
            return -WIN_SCORE;
        points += score
    
        # считаем фишки в восходящей диагонали:
        score = self.scorePosition(board, 0, 2, 1, -1)
        if (score == WIN_SCORE):
            return WIN_SCORE
        if (score == -WIN_SCORE):
            return -WIN_SCORE
        points += score
    
        return points
    
    
    # ПОДСЧИТЫВАЕМ ФИШКИ В ЗАДАННОМ РЯДУ
    def scorePosition(self, board, col, row, dcol, drow):
        # очки Игрока:
        playerPoints = 0
        # очки Компьютерa:
        computerPoints = 0
    
        for i in range(SIZE):
            if (board[col][row] == IPLAYER):
                playerPoints += 1
            elif (board[col][row] == ICOMPUTER):
                computerPoints += 1
            # координаты след. клетки ряда:
            col += dcol
            row += drow
    
        # выиграл Игрок:
        if (playerPoints == SIZE):
            return -WIN_SCORE
        # выиграл Компьютер:
        elif (computerPoints == SIZE):
            return WIN_SCORE
        else:
            # игра продолжается:
            return computerPoints
        
    
    # ПРОВЕРЯЕМ, НЕ ЗАПОЛНЕНА
    # ЛИ ДОСКА ФИШКАМИ
    def isFull(self, board):
        for row in range(self.rows):
            for col in range(self.cols):
                if (board[col][row] == IEMPTY):
                    return False
        return True
    

    # ВОЗВРАЩАЕМ КОПИЮ ДОСКИ board
    def boardCopy(self, board):
        return copy.deepcopy(board)
            
